from typing import Literal
from enum import Enum
from datetime import datetime
import time
from dataclasses import dataclass, field


class OutcomeType(Enum):
    BINARY = "BINARY"
    MULTIPLE_CHOICE = "MULTIPLE_CHOICE"
    PSEUDO_NUMERIC = "PSEUDO_NUMERIC"
    POLL = "POLL"
    BOUNTIED_QUESTION = "BOUNTIED_QUESTION"


class DescriptionFormat(Enum):
    HTML = "descriptionHTML"
    MARKDOWN = "descriptionMarkdown"
    JSON = "descriptionJSON"


@dataclass(kw_only=True)
class _CreateMarket:
    question: str
    closeTime: datetime = None
    description: str | tuple[str, DescriptionFormat] = None
    visibility: Literal["public", "unlisted"] = None
    groupIds: list[str] = None
    extraLiquidity: int = None

    def to_json(self):
        """
        Converts the create market request dataclass to a JSON dictionary suitable for sending as a request to the
        Manifold API, reformatting the close time and description as needed.

        Returns
        -------
        json: dict[str, Any]
            The JSON dictionary to be sent as a request.
        """
        json = {k: v for k, v in self.__dict__.items() if v is not None}
        if "outcomeType" in json:
            json["outcomeType"] = self.outcomeType.value
        if "closeTime" in json:
            json["closeTime"] = int(time.mktime(self.closeTime.timetuple()) * 1000)
        if "description" in json and isinstance(self.description, tuple):
            json[self.description[1].value] = self.description[0]
            del json["description"]

        return json


@dataclass
class CreateBinaryMarket(_CreateMarket):
    initialProb: int
    outcomeType: OutcomeType = field(default=OutcomeType.BINARY)


@dataclass
class CreatePseudoNumericMarket(_CreateMarket):
    question: str
    min: float
    max: float
    isLogScale: bool
    initialValue: float
    outcomeType: OutcomeType = field(default=OutcomeType.PSEUDO_NUMERIC)


@dataclass
class CreateMultipleChoiceMarket(_CreateMarket):
    question: str
    answers: list[str]
    addAnswersMode: Literal["DISABLED", "ONLY_CREATOR", "ANYONE"] = "DISABLED"
    outcomeType: OutcomeType = field(default=OutcomeType.MULTIPLE_CHOICE)
    shouldAnswersSumToOne: bool = True

    def to_json(self):
        dictionary = super().to_json()
        if (
            "shouldAnswersSumToOne" in dictionary
            and dictionary["shouldAnswersSumToOne"]
        ):
            del dictionary["shouldAnswersSumToOne"]
        return dictionary


@dataclass
class CreatePollMarket(_CreateMarket):
    answers: list[str]
    outcomeType: OutcomeType = field(default=OutcomeType.POLL)


@dataclass
class CreateBountiedQuestionMarket(_CreateMarket):
    totalBounty: int
    outcomeType: OutcomeType = field(default=OutcomeType.BOUNTIED_QUESTION)


CreateMarketRequest = (
    CreateBinaryMarket
    | CreatePseudoNumericMarket
    | CreateMultipleChoiceMarket
    | CreateBountiedQuestionMarket
    | CreatePollMarket
)


class RequestModel:
    def to_json(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}


@dataclass
class CreateBetRequest(RequestModel):
    amount: int
    contractId: str
    outcome: Literal["YES", "NO"] = field(default="YES")
    limitprob: float = None
    expiresAt: datetime = None

    def to_json(self):
        json = super().to_json()
        if "limitprob" in json:
            json["limitprob"] = round(json["limitprob"], 2)
        if "expiresAt" in json:
            json["expiresAt"] = int(time.mktime(self.expiresAt.timetuple()) * 1000)
        return json


@dataclass
class GetMarketsRequest(RequestModel):
    """
    Request parameters for the get_markets method, to list all markets, defaulting to creation date descending.

    Parameters
    ----------
    limit : int
        Optional. The maximum number of markets to return.
    sort : Literal["created-time", "updated-time", "last-bet-time", "last-comment-time"]
        Optional. The timestamp to sort by. Default is creation time.
    order : Literal["asc", "desc"]
        Optional. The order to sort by. Default is descending.
    before : str
        Optional. The ID of the market before which the list will start.
    userId : str
        Optional. Include only markets created by this user.
    groupId : str
        Optional. Include only markets tagged with this topic.
    """

    limit: int = None
    sort: Literal[
        "created-time", "updated-time", "last-bet-time", "last-comment-time"
    ] = None
    order: Literal["asc", "desc"] = None
    before: str = None
    userId: str = None
    groupId: str = None


@dataclass
class GetBetsRequest(RequestModel):
    userId: str = None
    username: str = None
    contractId: str = None
    contractSlug: str = None
    limit: int = None
    before: str = None
    after: str = None
    kinds: Literal["open-limit"] = None
    order: Literal["asc", "desc"] = None


@dataclass
class GetCommentsRequest(RequestModel):
    contractId: str = None
    contractSlug: str = None
    limit: int = None
    page: int = None
    userId: str = None


@dataclass
class SearchRequest(RequestModel):
    term: str
    sort: Literal["score", "newest", "liquidity"] = None
    filter: Literal[
        "all", "open", "closed", "resolved", "closing-this-month", "closing-next-month"
    ] = None
    contractType: Literal["ALL", "BINARY", "MULTIPLE_CHOICE", "BOUNTY", "POLL"] = None
    topicSlug: str = None
    creatorId: str = None
    limit: int = None
    offset: int = None


@dataclass
class GetGroupsRequest(RequestModel):
    beforeTime: datetime = None
    availableToUserId: str = None

    def to_json(self):
        json = super().to_json()
        if "beforeTime" in json:
            json["beforeTime"] = int(time.mktime(self.beforeTime.timetuple()) * 1000)
        return json


@dataclass
class GetPositionsRequest(RequestModel):
    order: Literal["shares", "profit"] = None
    top: int = None
    bottom: int = None
    userId: str = None


@dataclass
class GetUsersRequest(RequestModel):
    limit: int = None
    before: str = None


@dataclass
class AwardBountyRequest(RequestModel):
    amount: int
    commentId: str


@dataclass
class ModifyGroupRequest(RequestModel):
    groupId: str
    remove: bool = None


@dataclass
class ResolveBinaryMarket(RequestModel):
    outcome: Literal["YES", "NO", "MKT", "CANCEL"]
    probabilityInt: int = None


@dataclass
class MultipleChoiceResolution:
    answer: str
    pct: int


@dataclass
class ResolveMultipleChoiceMarket(RequestModel):
    outcome: Literal["MKT", "CANCEL"] | int
    resolutions: list[MultipleChoiceResolution] = None

    def to_json(self):
        json = super().to_json()
        if "resolutions" in json:
            if isinstance(self.resolutions, list):
                json["resolutions"] = [r.__dict__ for r in self.resolutions]
            else:
                del json["resolutions"]
        return json


@dataclass
class ResolveNumericMarket(RequestModel):
    """
    Parameters
    ----------
    outcome : Literal["CANCEL"] | int
        One of `CANCEL`, or a number indicating the selected numeric bucket ID.

    value : float
        The value that the market resolves to.

    probabilityInt : float
        Required if value is present. Should be equal to
        - If log scale: `log10(value - min + 1) / log10(max - min + 1)`
        - Otherwise: `(value - min) / (max - min)`
    """

    outcome: Literal["CANCEL"] | int
    value: float = None
    probabilityInt: float = None


ResolveMarketRequest = (
    ResolveBinaryMarket | ResolveMultipleChoiceMarket | ResolveNumericMarket
)


@dataclass
class SellSharesRequest(RequestModel):
    outcome: Literal["YES", "NO"] = None
    shares: int = None
    answerId: str = None


@dataclass
class SellSharesDPMRequest(RequestModel):
    contractId: str = None
    betId: str = None


@dataclass
class CreateCommentRequest(RequestModel):
    contractId: str
    description: str | tuple[str, Literal["content", "html", "markdown"]] = None

    def to_json(self):
        json = super().to_json()
        if "description" in json and isinstance(self.description, tuple):
            json[self.description[1]] = self.description[0]
            del json["description"]

        return json


@dataclass
class GetManagramsRequest(RequestModel):
    toId: str = None
    fromId: str = None
    limit: int = None
    before: datetime = None
    after: datetime = None

    def to_json(self):
        json = super().to_json()
        if "before" in json:
            json["before"] = int(time.mktime(self.before.timetuple()) * 1000)
        if "after" in json:
            json["after"] = int(time.mktime(self.after.timetuple()) * 1000)
        return json


@dataclass
class GetLeaguesRequest(RequestModel):
    userId: str = None
    season: int = None
    cohort: str = None



from datetime import datetime
from typing import Any
import requests
import logging

from .request_types import (
    AwardBountyRequest,
    CreateCommentRequest,
    CreateMarketRequest,
    GetBetsRequest,
    GetCommentsRequest,
    GetGroupsRequest,
    GetLeaguesRequest,
    GetManagramsRequest,
    GetMarketsRequest,
    GetPositionsRequest,
    GetUsersRequest,
    ModifyGroupRequest,
    ResolveMarketRequest,
    SearchRequest,
    CreateBetRequest,
    RequestModel,
    SellSharesDPMRequest,
    SellSharesRequest,
)


class Client:
    def __init__(self, api_key: str, api_url: str = "https://api.manifold.markets/v0"):
        """
        Creates a new Manifold client with a given API key.

        Parameters
        ----------
        api_key : str
            Your Manifold API key, which can be found in the settings on your profile.
        api_url : str, optional
            The URL of the Manifold API. Defaults to the current production Manifold API domain, "https://api.manifold.markets/v0".
        """
        self.api_url = api_url
        self.api_key = api_key

    def _get(
        self, endpoint: str, params: RequestModel = RequestModel()
    ) -> requests.Response:
        logging.info(f"GETting from {self.api_url}{endpoint} with {params.to_json()}")
        return requests.get(
            f"{self.api_url}{endpoint}",
            params=params.to_json(),
            headers={"Authorization": f"Key {self.api_key}"},
        )

    def _post(
        self, endpoint: str, request: RequestModel = RequestModel()
    ) -> requests.Response:
        logging.info(f"POSTing to {self.api_url}{endpoint} with {request.to_json()}")
        return requests.post(
            f"{self.api_url}{endpoint}",
            json=request.to_json(),
            headers={"Authorization": f"Key {self.api_key}"},
        )

    def get_user(self, username: str) -> requests.Response:
        return self._get(f"/user/{username}")

    def get_user_by_id(self, id: str) -> requests.Response:
        return self._get(f"/user/by-id/{id}")

    def get_me(self) -> requests.Response:
        return self._get("/me")

    def get_groups(
        self, request: GetGroupsRequest = GetGroupsRequest()
    ) -> requests.Response:
        return self._get("/groups", params=request)

    def get_group(self, slug: str) -> requests.Response:
        return self._get(f"/group/{slug}")

    def get_group_by_id(self, id: str) -> requests.Response:
        return self._get(f"/group/by-id/{id}")

    def get_markets(
        self, request: GetMarketsRequest = GetMarketsRequest()
    ) -> requests.Response:
        return self._get("/markets", params=request)

    def get_market(self, market_id: str) -> requests.Response:
        return self._get(f"/market/{market_id}")

    def get_positions(
        self, market_id: str, request: GetPositionsRequest
    ) -> requests.Response:
        return self._get(f"/market/{market_id}/positions", params=request)

    def get_market_by_slug(self, market_slug: str) -> requests.Response:
        return self._get(f"/slug/{market_slug}")

    def search_markets(self, request: SearchRequest) -> requests.Response:
        return self._get("/search-markets", params=request)

    def get_users(self, request: GetUsersRequest) -> requests.Response:
        return self._get("/users", params=request)

    def create_bet(self, request: CreateBetRequest) -> requests.Response:
        return self._post("/bet", params=request)

    def cancel_bet(self, id: str) -> requests.Response:
        return self._post(f"/bet/cancel/{id}")

    def create_market(self, market: CreateMarketRequest) -> requests.Response:
        """
        Posts a request to create a market on Manifold.

        Parameters
        ----------
        market : CreateMarketRequest
            The market to create. Can be of any of the types supported by Manifold: binary (0% - 100%), pseudo-numeric (minimum to maximum), multiple choice, bountied question, or a poll question.

        Returns
        -------
        requests.Response
            The response from the Manifold API.
        """
        return self._post("/market", market)

    def create_market_answer(self, market_id: str, text: str) -> requests.Response:
        return self._post(f"/market/{market_id}/answer", {"text": text})

    def add_liquidity(self, market_id: str, amount: int) -> requests.Response:
        return self._post(f"/market/{market_id}/add-liquidity", {"amount": amount})

    def add_bounty(self, market_id: str, amount: int) -> requests.Response:
        return self._post(f"/market/{market_id}/add-bounty", {"amount": amount})

    def award_bounty(
        self, market_id: str, award: AwardBountyRequest
    ) -> requests.Response:
        return self._post(f"/market/{market_id}/award-bounty", award)

    def set_close_time(
        self, market_id: str, close_time: datetime = None
    ) -> requests.Response:
        close_time_millis = int(close_time.timestamp() * 1000) if close_time else None
        return self._post(
            f"/market/{market_id}/close",
            {"closeTime": close_time_millis} if close_time else {},
        )

    def modify_group(
        self, market_id: str, request: ModifyGroupRequest
    ) -> requests.Response:
        return self._post(f"/market/{market_id}/group", request)

    def resolve_market(
        self, market_id: str, resolution: ResolveMarketRequest
    ) -> requests.Response:
        return self._post(f"/market/{market_id}/resolve", resolution)

    def sell_shares(
        self, market_id: str, request: SellSharesRequest
    ) -> requests.Response:
        return self._post(f"/market/{market_id}/sell", request)

    def sell_shares_dpm(
        self, market_id: str, request: SellSharesDPMRequest
    ) -> requests.Response:
        return self._post(f"/sell-shares-dpm", request)

    def create_comment(self, comment: CreateCommentRequest) -> requests.Response:  #
        return self._post("/comment", comment)

    def get_comments(
        self, request: GetCommentsRequest = GetCommentsRequest()
    ) -> requests.Response:
        return self._get("/comments", params=request)

    def get_bets(self, request: GetBetsRequest = GetBetsRequest()) -> requests.Response:
        return self._get("/bets", params=request)

    def get_managrams(
        self, request: GetManagramsRequest = GetManagramsRequest()
    ) -> requests.Response:
        return self._get("/managrams", params=request)

    def get_leagues(
        self, request: GetLeaguesRequest = GetLeaguesRequest()
    ) -> requests.Response:
        return self._get("/leagues", params=request)
